/// fike-1.0

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/ioctl.h>
#include <net/if.h>    
#include <sys/select.h>
#include "fike.h"

#ifndef SOCK_CLOEXEC
#ifdef O_CLOEXEC
#define SOCK_CLOEXEC O_CLOEXEC
#else
#define SOCK_CLOEXEC 0
#endif
#endif 

#define FIKE_MULTICAST_ADDRESS      "239.1.4.73"
#define FIKE_MULTICAST_PORT         4975

// these variables are the same values, but in network byte order
static uint32_t fike_multicast_address;
static uint32_t fike_multicast_port;

static void fike_translate_constants()
{
  if (fike_multicast_address) return ;
  fike_multicast_address= inet_addr(FIKE_MULTICAST_ADDRESS);
  fike_multicast_port= htons(FIKE_MULTICAST_PORT);
}


// magic for commands
#define FIKE_MAGIC     0xd3614afe

// magic for directory search query
#define FIKE_QUERY     0xfa357168
// and its response
#define FIKE_RESPONSE  0xaf351768

// directory search/response message length
#define FIKE_MSGLEN    50

// directory search message format
#define FIKE_mMAGIC    0
#define FIKE_mHOST     4
#define FIKE_mPORT     8
#define FIKE_mTYPE     10
#define FIKE_mNAME     12
#define FIKE_NAMELEN   (FIKE_MSGLEN-FIKE_mNAME)

#define FIKE_BIT31     0x80000000

#ifdef fike_self_test
static void fike_print_message(unsigned char *,int); 
#endif

/***********************************************************************
  There are many types of sockets in here. I should give them some 
  names so that I don't confuse myself with server, service, etc.

  Server side:
     kabul:    is used to accept connections from clients who have
               already discovered us.
     sesver:   is used to respond to directory queries. when we
               publish a service, this socket is put into the multicast
               group.
     kart:     it refers to a network interface.
     musteri:  this socket is the result of accepting a connection to
               kabul.
 ***********************************************************************/


/***********************************************************************
                          Error Handling
 ***********************************************************************/
void (*fike_error_output)(const char *data, int len);

static void fike_error_stderr(const char *data,int len)
{
  fwrite(data,1,len,stderr);
}

static void fike_error(const char *fmt,...)
{
  va_list ap;
  char buf[2000];
  char ibuf[20];
  int Vi;
  char *Vs;
  int o,s,L;
  int eno;

  va_start(ap,fmt);
  if (!fike_error_output) fike_error_output= fike_error_stderr;
  o= 0; s= sizeof(buf)-2;

  eno= errno;
  while(s && *fmt)
  {
    if (fmt[0]!='%') { buf[o++]= *fmt++; s--; continue; }
    fmt++;
    Vs= NULL;
    switch(fmt[0])
    {
    case 'd': Vi= va_arg(ap,int); sprintf(ibuf,"%d",Vi); Vs= ibuf; break;
    case 'e': Vs= strerror(eno); break;
    case 's': Vs= va_arg(ap,char*); break;
    case 0: fmt--; break;
    default: ibuf[0]= '%'; ibuf[1]= fmt[0]; ibuf[2]= 0; Vs= ibuf; break;
    }
    if (!Vs) break;
    fmt++;
    L= strlen(Vs);
    if (L>s) L= s;
    memcpy(buf+o, Vs, L);
    o+= L;
    s-= L;
  }
  va_end(ap);
  buf[o++]= '\n';
  buf[o++]= 0;
  (*fike_error_output)(buf,o);
}

#define FIKE_ERROR(label, fmt, ...)  \
   do { fike_error(fmt, ##__VA_ARGS__ ); goto label; } while (0)

/***********************************************************************
                    POINTER ARRAYS
   data is in fact a triple pointer, but for sake of 
   avoiding useless warnings, I'll make it a single one
 ***********************************************************************/
typedef struct { void *data; int *n_data; int s_data; } FIKE_PARRAY; 

static void fike_parray_add(FIKE_PARRAY *PA, void *ptr)
{
  void ***DD;
  DD= PA->data;
  if (!DD) return ; // uninitialized
  if (PA->n_data[0]==PA->s_data)
  {
    PA->s_data += PA->s_data/2 + 10;
    *DD= realloc(*DD, PA->s_data*sizeof(void*));
  }
  (*DD)[PA->n_data[0]++]= ptr;
}

static void fike_parray_remove_value(FIKE_PARRAY *PA, void *value)
{
  void ***DD;
  int i;

  DD= PA->data;
  if (!DD) return ;
  
  for(i=0;i<PA->n_data[0];i++)
    if ((*DD)[i]==value)
    {
      (*DD)[i]= (*DD)[PA->n_data[0]-1];
      PA->n_data[0]--;
      return ;
    }
}

/***********************************************************************
   ZABIT:  a record of active FDs
 ***********************************************************************/
enum fike_zabit_types 
  { FIKE_ZABIT_SESVER, FIKE_ZABIT_KABUL, FIKE_ZABIT_MUSTERI,
    FIKE_ZABIT_QUERY, FIKE_ZABIT_SERVER };

typedef struct 
{
  int fd;
  void *data;
  int type; 
} FIKE_ZABIT;

static FIKE_ZABIT **fike_zabit; static int fike_n_zabit;
static FIKE_PARRAY fike_zabit_a= { &fike_zabit, &fike_n_zabit };

static FIKE_ZABIT* fike_create_zabit(int type, void *data, int fd)
{
  FIKE_ZABIT *FR;
  FR= calloc(1,sizeof(*FR));
  FR->type= type;
  FR->data= data;
  FR->fd= fd;
  fike_parray_add(&fike_zabit_a, FR);
  return FR;
}



typedef struct 
{
  unsigned char obuffer[65536+10];
  unsigned char *optr;
  unsigned char ibuffer[65536+10];
  unsigned char *ibufferend, *iptr;
  int fd;
} FIKE_IO;

/***********************************************************************
  Network card code. Each interface is represented by a FIKE_KART.
  If the service uses a given card, we make a sesver socket for that
  kart. On the client side, this struct is used to select which card
  to use for query. After the query, connection is made without reference
  to any network card.
 ***********************************************************************/

typedef struct
{
  char *iface;
  int sesver;
  int kabul;
  uint32_t iface_addr;
} FIKE_KART;

static FIKE_KART **fike_kart; static int fike_n_kart;
static FIKE_PARRAY fike_kart_a= { &fike_kart, &fike_n_kart };

static void fike_init_kart(char *iface)
{
  FIKE_KART *kart;
  kart= calloc(1,sizeof(*kart)); 
  kart->iface= strdup(iface);
  fike_parray_add(&fike_kart_a, kart);
  kart->iface_addr= fike_get_network_interface_address(iface); 
  kart->sesver= -1;
  kart->kabul= -1;
}

static FIKE_KART* fike_get_kart(char *iface)
{
  int i;
  for(i=0;i<fike_n_kart;i++)
   if (!strcmp(iface, fike_kart[i]->iface)) 
   {
     if (fike_kart[i]->iface_addr!=INADDR_ANY) return fike_kart[i];
                                          else break;
   }
  return NULL;
}

static void fike_init_all_kart()
{
  char **ifaces;
  int i;
  if (fike_n_kart) return ;
  ifaces= fike_get_network_interfaces(1);
  for(i=0;ifaces[i];i++)
    fike_init_kart(ifaces[i]);
  fike_free_network_interfaces(ifaces);
}


#ifdef FIKE_IMP_SERVER

struct fike_client
{
  char host[50];
  int port;
  void *usr;
  FIKE_IO io;
  int cmd;
};

static int fike_service_port, fike_service_type;
static char *fike_service_name;
#endif


#ifdef FIKE_IMP_CLIENT
struct fike_server
{
  int port; uint32_t host;
  FIKE_IO io;
};

static FIKE_SERVER **fike_server; static int fike_n_server;
static FIKE_PARRAY fike_server_a= {&fike_server, &fike_n_server};
#endif


static void fike_close_sock(int fd)
{
  fike_remove_fd(fd);
  shutdown(fd,2);
  close(fd);
}



uint32_t fike_get_network_interface_address(char *iface)
{
  int sock;
  uint32_t R;
  struct ifreq req;
  struct sockaddr_in *inaddr;

  R= INADDR_ANY;
  sock= socket(AF_INET, SOCK_DGRAM, 0);
  if (sock<0) 
    FIKE_ERROR(err_sock,"can't create socket: %e");

  strcpy(req.ifr_name, iface);
  if (ioctl(sock, SIOCGIFADDR, &req)<0) 
    FIKE_ERROR(err_ioctl,"can't get address for iface %s: %e", iface);

  inaddr= (struct sockaddr_in*) &req.ifr_addr;
  R= inaddr->sin_addr.s_addr;
err_ioctl:
  close(sock);
err_sock:
  return R;              
}

char** fike_get_network_interfaces(int filter_loopback)
{
  int sock;
  struct ifreq reqs[20];
  struct ifconf conf;
  char **R;
  int n;
  int i,j;

  if ((sock= socket(AF_INET, SOCK_DGRAM, 0))<0) 
     FIKE_ERROR(Enoclose, "socket(): %e"); 

  conf.ifc_len= sizeof(reqs);
  conf.ifc_req= reqs;
  if (ioctl(sock, SIOCGIFCONF, &conf)<0) 
     FIKE_ERROR(Eclose, "ioctl(): %e"); 
  close(sock);

  n= conf.ifc_len/sizeof(reqs[0]);
  R= malloc(sizeof(*R)*(n+1));
  for(i=0,j=0;i<n;i++) 
  {
    if (filter_loopback && reqs[i].ifr_name[0]=='l' &&
        reqs[i].ifr_name[1]=='o') continue;
    R[j++]= strdup(reqs[i].ifr_name);
  }
  R[j]= NULL;
  return R;
Eclose:
  close(sock);
Enoclose:
  R= malloc(1*sizeof(*R));
  R[0]= NULL;
  return R;
}

void fike_free_network_interfaces(char **K)
{
  int i;
  for(i=0;K[i];i++) free(K[i]);
  free(K);
}


#ifdef FIKE_IMP_SERVER

static FIKE_CLIENT **fike_client; static int fike_n_client;
static FIKE_PARRAY fike_client_a= { &fike_client, &fike_n_client };

static int fike_open_sesver(uint32_t iface_addr)
{
  int fd;
  struct ip_mreqn mg;
  struct sockaddr_in addr;
  struct linger L;
  int zero, one;

  zero= 0;
  one= 1;
  fike_translate_constants();

  if ((fd= socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0))<0)
     FIKE_ERROR(Enoclose, "socket(): %e");

  setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void*) &one, sizeof(one));
#ifdef SO_REUSEPORT
  setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (void*) &one, sizeof(one));
#endif
  L.l_onoff = 0;
  L.l_linger = 0;
  setsockopt(fd, SOL_SOCKET, SO_LINGER, (void*) &L, sizeof(L));
                                                     
  memset(&addr, 0, sizeof(addr));
  addr.sin_family= AF_INET;
  /// shouldn't this be iface_addr??  Maybe not: see
  ///  https://stackoverflow.com/questions/12681097/c-choose-interface-for-udp-multicast-socket
  addr.sin_addr.s_addr= htonl(INADDR_ANY);
  addr.sin_port= fike_multicast_port;
  if (bind(fd,(struct sockaddr *)&addr, sizeof(addr))<0) 
     FIKE_ERROR(Eclose, "bind(): %e");
                                        
  mg.imr_multiaddr.s_addr= fike_multicast_address;
  mg.imr_address.s_addr= iface_addr;
  mg.imr_ifindex= 0;
  if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*) &mg, sizeof(mg)))
     FIKE_ERROR(Eclose, "addmulticast(): %e");

  if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&zero, sizeof(zero)))
     FIKE_ERROR(Eclose,"disableloop: %e");

  fike_add_fd(fd);
  return fd;
Eclose:
  close(fd);
Enoclose:
  return -1;
}

static int fike_open_kabul(int port,uint32_t iface_addr)
{
  int one;
  struct sockaddr_in addr;
  struct linger L;
  int fd;

  one= 1;

  fd= socket(AF_INET, SOCK_STREAM | SOCK_CLOEXEC, 0);
  if (fd<0) 
     FIKE_ERROR(Enoclose, "server.socket(): %e");

  setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void*) &one, sizeof(one));
#ifdef SO_REUSEPORT
  setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, (void*) &one, sizeof(one));
#endif
  L.l_onoff = 0;
  L.l_linger = 0;
  setsockopt(fd, SOL_SOCKET, SO_LINGER, (void*) &L, sizeof(L));

  memset(&addr, 0, sizeof(addr));
  addr.sin_family= AF_INET;
  addr.sin_addr.s_addr= iface_addr;
  addr.sin_port= htons(port);
  if (bind(fd,(struct sockaddr *)&addr, sizeof(addr))<0) 
     FIKE_ERROR(Eclose, "server.bind(): %e");

  if (listen(fd, 5)<0) 
     FIKE_ERROR(Eclose, "server.listen(): %e");

  signal(SIGPIPE, SIG_IGN);
  fike_add_fd(fd);
  return fd;
Eclose:
  close(fd);
Enoclose:
  return -1;
}

static int fike_put_service_on_kart(FIKE_KART *kart)
{
  if (kart->sesver==-1)
  {
    if (kart->iface_addr==INADDR_ANY) return 1;
    kart->sesver= fike_open_sesver(kart->iface_addr);
    if (kart->sesver==-1) return 1;
    fike_create_zabit(FIKE_ZABIT_SESVER, kart, kart->sesver);
  }
  kart->kabul= fike_open_kabul(fike_service_port, kart->iface_addr);
  if (kart->kabul<0) return 1;
  fike_create_zabit(FIKE_ZABIT_KABUL, kart, kart->kabul);
  return 0;
}

int fike_publish(char *iface,char *name, int type, int port)
{
  FIKE_KART *kart;
  int R;
  int i;

  fike_translate_constants();
  fike_init_all_kart();

  if (!name) FIKE_ERROR(Enofree, "internal error: must give a name"); 

  fike_service_name= strdup(name); 
  fike_service_type= type;
  fike_service_port= port;
  if (iface)
  {
    kart= fike_get_kart(iface);
    R= fike_put_service_on_kart(kart);
  }
  else
  {
    R= 1;
    for(i=0;i<fike_n_kart;i++)
      R *= fike_put_service_on_kart(fike_kart[i]);
  }
  if (R) FIKE_ERROR(Enofree, "no server sockets could be created");
  return 0;

Enofree:
  return 1;
}
#endif // IMP_SERVER

int fike_n_fds() { return fike_n_zabit; }
int fike_fd(int idx) 
{
  if (idx<0 || idx>=fike_n_zabit) return 0;
  return fike_zabit[idx]->fd;
}

int fike_get_select(fd_set *rset, fd_set *xset)
{
  int i,fd, max;
  FIKE_ZABIT *Z;
  max= -1;
  for(i=0;i<fike_n_zabit;i++)
  {
    fd= fike_zabit[i]->fd;
    if (fd<0)
    {
      Z= fike_zabit[i];
      fike_zabit[i]= fike_zabit[fike_n_zabit-1];
      fike_n_zabit--;
      i--;
      free(Z);
      continue;
    }
    if (rset) FD_SET(fd, rset);
    if (xset) FD_SET(fd, xset);
    if (fd>max) max= fd;
  }
  return max;
}

static int fike_recv_force(int fd, unsigned char *buf, int len)
{
  int R;
  if (fd<0) return 1;
  while(len)
  {
    R= recv(fd, buf, len, 0);
    if (R<0 && errno==EINTR) continue;
    if (R<0) FIKE_ERROR(recv_error, "recv: %e");
    if (R==0) return 1;  // orderly shutdown
    buf += R;
    len -= R;
  }
  return 0;
recv_error:
  return 1;
}

static int fike_get_int(FIKE_IO *io,int *R)
{
  if (io->iptr+4>io->ibufferend) return 1;
  *R=  io->iptr[0] | (io->iptr[1]<<8) | (io->iptr[2]<<16) | (io->iptr[3]<<24);
  io->iptr+= 4;
  return 0;
}

static int fike_get_word(FIKE_IO *io,unsigned short *R)
{
  if (io->iptr+2>io->ibufferend) return 1;
  *R=  io->iptr[0] | (io->iptr[1]<<8);
  io->iptr+= 2;
  return 0;
}

static int fike_recv_msg(FIKE_IO *io, int *cmd)
{
  unsigned short sz;
  int magic;

  if (io->fd<0) return 1;
  if (fike_recv_force(io->fd, io->ibuffer, 10)) return 1;
  io->ibufferend= io->ibuffer+10;

  io->iptr= io->ibuffer; 
  fike_get_int(io,&magic);
  if (magic!=FIKE_MAGIC) FIKE_ERROR(bad_magic, "Bad magic in message");
  fike_get_int(io, cmd);    // get the command
  fike_get_word(io, &sz);       // get the size

  io->ibufferend= io->ibuffer+ 10 + sz;
  if (fike_recv_force(io->fd, io->ibuffer+10, sz)) return 1;

  return 0;
bad_magic:
  return 1;
}

static void fike_process_zabit_x(FIKE_ZABIT *Z)
{
#ifdef FIKE_IMP_SERVER
  FIKE_KART *kart;
  FIKE_CLIENT *cc;
#endif
  switch(Z->type)
  {
  case FIKE_ZABIT_KABUL:
#ifdef FIKE_IMP_SERVER
    fike_close_sock(Z->fd);
    kart= Z->data;
    kart->kabul= -1;
#endif
    break;
  case FIKE_ZABIT_SESVER:
#ifdef FIKE_IMP_SERVER
    kart= Z->data;
    fike_close_sock(Z->fd);
    kart->sesver= -1;
#endif
    break;
  case FIKE_ZABIT_MUSTERI:
#ifdef FIKE_IMP_SERVER
    cc= Z->data;
    fike_close_sock(cc->io.fd);
    cc->io.fd= -1;
    fike_client_disconnected(cc->host, cc->port, cc->usr);
    fike_parray_remove_value(&fike_client_a, cc);
#endif
    break;
  case FIKE_ZABIT_QUERY:
#ifdef FIKE_IMP_CLIENT
    fike_close_sock(Z->fd);
#endif
    break;
  } 
  Z->fd= -1;
}

#ifdef FIKE_IMP_SERVER
static int fike_accept(FIKE_KART* kart, FIKE_CLIENT *cc)
{
  struct sockaddr_in addr;
  socklen_t alen;

  if (kart->kabul<0) FIKE_ERROR(Enoclose, "bad socket to accept from");

  memset(&addr, 0, sizeof(addr));
  alen= sizeof(addr);
printf("Accepting from socket %d\n", kart->kabul);
  cc->io.fd= accept(kart->kabul, (struct sockaddr*) &addr, &alen);
printf("Accept result: %d\n", cc->io.fd);

  //
  // fixme: disable the socket if the error is a permanent
  // one, like errno==ENOTSOCK etc.
  //
  if (cc->io.fd<0) FIKE_ERROR(Enoclose, "accept failed: %e");

  fike_add_fd(cc->io.fd);
  fike_create_zabit(FIKE_ZABIT_MUSTERI,cc,cc->io.fd);

  inet_ntop(AF_INET,&addr.sin_addr,cc->host,sizeof(cc->host));
  cc->port= ntohs(addr.sin_port);
  return 0;
Enoclose:
  return 1;
}
#endif

static uint32_t fike_gint32(unsigned char *B)
{
  int i;
  uint32_t V= 0;
  for(i=0;i<4;i++) { V= V<<8; V |= B[i]; }
  return V;
}

static uint32_t fike_gint16(unsigned char *B)
{
  int i;
  uint32_t V= 0;
  for(i=0;i<2;i++) { V= V<<8; V |= B[i]; }
  return V;
}

static void fike_pint32(unsigned char *B, uint32_t V)
{
  int i;
  for(i=0;i<4;i++) { B[i]= V>>24; V<<= 8; }
}

static void fike_pint16(unsigned char *B, uint16_t V)
{
  int i;
  for(i=0;i<2;i++) { B[i]= V>>8; V<<= 8; }
}

#ifdef FIKE_IMP_SERVER
static int fike_pmatch(char *P, char *V)
{
  char *z;
again:
  if (P[0]!='*')
  {
    if (P[0]!=V[0]) return 0;
    if (P[0]==0) return 1;
    P++; V++; goto again;
  }
  if (P[1]==0) return 1;
  if (P[1]=='*') { P++; goto again; }

  for(z=V;z[0];z++) if (z[0]==P[1]  && fike_pmatch(P+1,z)) return 1;
  return 0;
}

static void fike_respond_to_query(FIKE_KART *kart)
{
  int rl, sl;
  struct sockaddr_in  sender; socklen_t slen;
  unsigned char buf[FIKE_MSGLEN+1]; 

  if (kart->sesver<0) return ;
  slen= sizeof(sender);
  memset(&sender, 0, slen);
  memset(buf, 0, sizeof(buf));
  rl= recvfrom(kart->sesver, buf, FIKE_MSGLEN, MSG_WAITALL,
               (struct sockaddr*)&sender, &slen);
  buf[FIKE_MSGLEN]= 0;

  if (rl!=FIKE_MSGLEN)
  {
    if (rl<0)
       FIKE_ERROR(sleepy, "recvfrom failed: %e");
    else
       FIKE_ERROR(sleepy,"short message recived: length= %d", rl);
  }

#ifdef fike_self_test
  fike_print_message(buf,rl);
#endif

  if (fike_gint32(buf + FIKE_mMAGIC)!=FIKE_QUERY) return ;
  
  if (!fike_pmatch((char*) buf+FIKE_mNAME, fike_service_name)) return ;
  memcpy(buf+FIKE_mNAME, fike_service_name, FIKE_NAMELEN);
  buf[FIKE_MSGLEN]= 0;

  fike_pint32(buf+FIKE_mMAGIC, FIKE_RESPONSE);
  fike_pint32(buf+FIKE_mHOST, kart->iface_addr);
  fike_pint16(buf+FIKE_mPORT, fike_service_port);
  fike_pint16(buf+FIKE_mTYPE, fike_service_type);

  usleep( (rand()%200)*1000 );

  sl= sendto(kart->sesver, buf, FIKE_MSGLEN, MSG_NOSIGNAL, 
             (struct sockaddr*) &sender, sizeof(sender));
  if (sl!=FIKE_MSGLEN)
  {
    if (sl<0)
      FIKE_ERROR(sleepy,"sendto failed: %e");
    else
      FIKE_ERROR(sleepy,"sent message was truncated: length= %d",sl);
  }
  return ;
sleepy:
  usleep(1000);
}
#endif

#ifdef FIKE_IMP_CLIENT
static void fike_read_query_result(FIKE_KART *kart, int fd)
{
  struct sockaddr_in  sender; socklen_t slen;
  unsigned char buf[FIKE_MSGLEN+1]; 
  int rl;
  uint32_t port, type;
  char *name;

  if (fd<0) return ;

  slen= sizeof(sender);
  memset(&sender, 0, slen);
  memset(buf, 0, sizeof(buf));
  rl= recvfrom(fd, buf, FIKE_MSGLEN, MSG_WAITALL,
               (struct sockaddr*)&sender, &slen);
  buf[FIKE_MSGLEN]= 0;
  if (rl!=FIKE_MSGLEN)
  {
    if (rl<0)
      FIKE_ERROR(fail,"recvfrom failed: %e");
    else
      FIKE_ERROR(fail,"short message received: length= %d", rl);
  }
#ifdef fike_self_test
  fike_print_message(buf,rl);
#endif
  if (fike_gint32(buf+FIKE_mMAGIC)!=FIKE_RESPONSE) return ;
  struct in_addr aa;
  aa.s_addr= fike_gint32(buf+FIKE_mHOST);
  port= fike_gint16(buf+FIKE_mPORT);
  type= fike_gint16(buf+FIKE_mTYPE);
  name= (char*) buf+FIKE_mNAME;
  fike_found_server(inet_ntoa(aa), aa.s_addr, port, type, name);
  return ;
fail:
  usleep(1000);
}
#endif

static void fike_process_zabit_r(FIKE_ZABIT *Z)
{
#ifdef FIKE_IMP_SERVER
  FIKE_CLIENT *cc;
#endif
  switch(Z->type)
  {
#ifdef FIKE_IMP_SERVER
  case FIKE_ZABIT_KABUL:
    cc= calloc(1,sizeof(*cc));
    if (fike_accept(Z->data, cc)) { free(cc); break; }
    cc->usr= fike_client_connected(cc->host, cc->port, cc->io.fd);
    if (!cc->usr) { fike_close_client(cc); break; }
    fike_parray_add(&fike_client_a, cc);
    break;
  case FIKE_ZABIT_SESVER:
    fike_respond_to_query(Z->data);
    break;
  case FIKE_ZABIT_MUSTERI:
    cc= Z->data;
    if (fike_recv_msg(&cc->io, &cc->cmd)) break;
    fike_client_message(cc, cc->usr, cc->cmd);
    break;
#endif
#ifdef FIKE_IMP_CLIENT
  case FIKE_ZABIT_QUERY:
    fike_read_query_result(Z->data, Z->fd);
    break;
#endif
  } 
  return ;
}

void fike_process_select(fd_set *rset, fd_set *xset)
{
  int i;
  FIKE_ZABIT *z; int fd;
  for(i=0;i<fike_n_zabit;i++)
  {
    z= fike_zabit[i]; fd= z->fd;
    if (FD_ISSET(fd, xset)) 
    {
      FD_CLR(fd, rset);
      fike_process_zabit_x(z);  
      FD_CLR(fd, xset);
    }
    if (FD_ISSET(fd, rset))
    {
      fike_process_zabit_r(z);
      FD_CLR(fd, rset);
    }
  }
}

#ifdef FIKE_IMP_CLIENT
static int fike_query_socket(uint32_t iface)
{
  struct in_addr li;
  int fd;

  fd= socket(AF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
  if (fd<0) FIKE_ERROR(Enoclose, "socket failed:%e");

#if 0
  flag= 0;
  call= "disableloop";
  if (setsockopt(fd, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&flag, sizeof(flag)))
    goto Fclose;
#endif

  li.s_addr= iface; 
  if (setsockopt(fd,IPPROTO_IP,  IP_MULTICAST_IF, (void*)&li, sizeof(li)))
    FIKE_ERROR(Eclose,"set local iface failed: %e");
  fike_add_fd(fd);
  return fd;
Eclose:
  close(fd);
Enoclose:
  return -1;
}

static int fike_query_on_kart(FIKE_KART *kart, char *pattern)
{
  struct sockaddr_in gaddr;
  int sl,q_fd;
  unsigned char buf[FIKE_MSGLEN+1];

  q_fd= fike_query_socket(kart->iface_addr);
  if (q_fd<0) return 1;

  memset(&gaddr, 0, sizeof(gaddr));
  gaddr.sin_family= AF_INET;
  gaddr.sin_addr.s_addr= fike_multicast_address;
  gaddr.sin_port= fike_multicast_port;

  memset(buf, 0, sizeof(buf));
  fike_pint32(buf+FIKE_mMAGIC, FIKE_QUERY);
  fike_pint32(buf+FIKE_mHOST, 0);
  fike_pint16(buf+FIKE_mPORT, 0);
  fike_pint16(buf+FIKE_mTYPE, 0);

  strncpy((char*) buf+FIKE_mNAME, pattern, FIKE_NAMELEN);
  buf[FIKE_MSGLEN]= 0;

  sl= sendto(q_fd, buf, FIKE_MSGLEN, MSG_NOSIGNAL,
             (struct sockaddr*) &gaddr, sizeof(gaddr));

  if (sl!=FIKE_MSGLEN)
  {
    if (sl<0)
       FIKE_ERROR(Eclose,"sendto failed: %e");
    else
       FIKE_ERROR(Eclose,"sent message was truncated: length= %d", sl);
  }

  fike_add_fd(q_fd);
  fike_create_zabit(FIKE_ZABIT_QUERY, kart, q_fd);
  return 0;

Eclose:
  close(q_fd);
  return 1;
}

int fike_start_query(char *iface,char *pattern)
{
  FIKE_KART *kart;
  int i;
  int R;

  fike_translate_constants();
  fike_init_all_kart();
  R= 0;
  if (!iface)
  {
    for(i=0;i<fike_n_kart;i++)
      R |= fike_query_on_kart(fike_kart[i], pattern);
  }
  else
  {
    kart= fike_get_kart(iface);
    if (kart) return 1;
    R= fike_query_on_kart(kart, pattern);
  }
  return R;
}

void fike_stop_query()
{
  int i;
  for(i=0;i<fike_n_zabit;i++)
    if (fike_zabit[i]->type==FIKE_ZABIT_QUERY)
    {
      fike_close_sock(fike_zabit[i]->fd);
      fike_zabit[i]->fd= -1;
    }
}


FIKE_SERVER* fike_connect(uint32_t addr, int port)
{
  struct sockaddr_in saddr;
  int fd;
  FIKE_SERVER *s;

  fd= socket(AF_INET, SOCK_STREAM, 0);
  if (fd<0) FIKE_ERROR(Enoclose, "can't create socket: %e");

  memset(&saddr, 0, sizeof(saddr));
  saddr.sin_family= AF_INET;
  saddr.sin_addr.s_addr= addr;
  saddr.sin_port= htons(port);
  if (connect(fd, (struct sockaddr*) &saddr, sizeof(saddr))<0)
    FIKE_ERROR(Eclose,"can't connect: %e");

  s= calloc(1,sizeof(*s));
  s->io.fd= fd;
  s->port= port;
  s->host= addr;
  fike_parray_add(&fike_server_a, s);
  fike_add_fd(fd);
  return s;

Eclose:
  close(fd);
Enoclose:
  return NULL;
}
#endif

static void fike_put_int(FIKE_IO* io,int V)
{
  int i;
  for(i=0;i<4;i++) {  *io->optr++= V & 0xff; V>>= 8; }
}

static void fike_put_word(FIKE_IO* io,int V)
{
  int i;
  for(i=0;i<2;i++) {  *io->optr++= V & 0xff; V>>= 8; }
}

static int fike_send_force(FIKE_IO *io, unsigned char *B, int L)
{
  int r;
  while(L)
  {
    r= send(io->fd, B, L, MSG_NOSIGNAL);
    if (r>=0) { B+= r; L-=r; continue; }
    if (r==-1 && errno==EINTR) continue;
    return 1;
  } 
  return 0;
}

static int fike_flush_output(FIKE_IO *io)
{
  if (io->optr==io->obuffer) return 0;
  if (fike_send_force(io, io->obuffer, io->optr-io->obuffer)) return 1;
  io->optr= io->obuffer;
  return 0;
}

static int fike_send_msgv(FIKE_IO *io, int cmd, const char *fmt, va_list ap)
{
  unsigned int size;
  va_list sap;
  int i;
  unsigned char *B; int L;

  if (io->fd<0) return 1;

  size= 0;
  va_copy(sap, ap);
  for(i=0;fmt[i];i++)
    switch(fmt[i])
    {
    case 'i': size+= 4; va_arg(sap,int); break;
    case 'w': size+= 2; va_arg(sap,int); break;
    case 'B': size+= 2; va_arg(sap, unsigned char*); 
              size+= va_arg(sap,int); break;
    case 's': size+= 2;
              B= va_arg(sap, unsigned char*);
              size+= B ? (strlen((char*)B)+1) : 1; break;
    }
  va_end(sap);

  if (size>65535) FIKE_ERROR(fail, "message too big size= %d",size);

  io->optr= io->obuffer;
  fike_put_int(io,FIKE_MAGIC);
  fike_put_int(io,cmd);
  fike_put_word(io,size);

  for(i=0;fmt[i];i++)
    switch(fmt[i])
    {
    case 'i': fike_put_int(io,va_arg(ap,int)); break;
    case 'w': fike_put_word(io,va_arg(ap,int)); break;
    case 'B': 
      B= va_arg(ap, unsigned char*);
      L= va_arg(ap, int);
      fike_put_word(io,L);
      if (fike_flush_output(io) ||fike_send_force(io,B,L)) return 1;
      break;
    case 's':
      B= va_arg(ap, unsigned char*);
      L= strlen((char*)B)+1;
      fike_put_word(io,L);
      if (fike_flush_output(io) || fike_send_force(io,B,L)) 
         return 1;
      break;
    default: 
      FIKE_ERROR(fail, "bad format char (%c)", *(fmt-1)); 
    }

  if (fike_flush_output(io)) return 1;
  return 0;

fail:
  return 1;
}

#ifdef FIKE_IMP_CLIENT
int fike_command
  (FIKE_SERVER *server, int response, int cmd, const char* fmt, ...)
{
  va_list ap;
  int R;
  int rc;

  va_start(ap,fmt);
  R= fike_send_msgv(&server->io, cmd, fmt, ap);
  va_end(ap);
  if (R) return 1;
  if (!response) return 0;

  if (fike_recv_msg(&server->io,&rc)) return 1; 
  if (rc!= (cmd|FIKE_BIT31))
     FIKE_ERROR(bad_response,"expected a response to %08x, got %08x\n",
                             cmd, rc); 
  return 0;
bad_response: 
  return 1;
}
#endif

#ifdef FIKE_IMP_SERVER
int fike_response(FIKE_CLIENT *cc, const char* fmt, ...)
{
  va_list ap;
  int R;
  va_start(ap,fmt);
  R= fike_send_msgv(&cc->io, cc->cmd|FIKE_BIT31, fmt, ap);
  va_end(ap);
  return R;
}
#endif

static int fike_get_params(FIKE_IO *io,const char *fmt, va_list ap)
{
  int R;
  int *Rl; unsigned char **Rd;
  unsigned short l;

  R= 1;
  while(*fmt)
    switch(*fmt++)
    {
    case 'i': 
      if (fike_get_int(io,va_arg(ap,int*))) goto toosmall; 
      break;
    case 'w':
      if (fike_get_word(io,va_arg(ap,unsigned short*))) goto toosmall;
      break;
    case 'B': 
      if (fike_get_word(io,&l)) goto toosmall;
      Rd= va_arg(ap, unsigned char**);
      Rl= va_arg(ap, int*);
      if (l+io->iptr>io->ibufferend) goto toosmall;
      *Rl= l;
      *Rd= io->iptr;
      io->iptr+= l;
      break;     
    case 's':
      if (fike_get_word(io,&l)) goto toosmall;
      Rd= va_arg(ap, unsigned char**);
      if (l+io->iptr>io->ibufferend) goto toosmall;
      *Rd= io->iptr;
      if ((*Rd)[l-1]) goto bad_format;
      io->iptr+= l;
      break;     
    default:
      goto bad_format;
    }
  R= 0;
done:
  return R;
bad_format: FIKE_ERROR(done,"bad message format");
toosmall:   FIKE_ERROR(done,"message too short"); 
}

static int fike_get_rest(FIKE_IO *io,unsigned char **Rd, unsigned int *Rl)
{
  if (io->ibufferend<io->iptr) return 1;
  *Rd= io->iptr;
  *Rl= io->ibufferend-io->iptr;
  return 0;
}

#ifdef FIKE_IMP_SERVER
int fike_command_params(FIKE_CLIENT *cc, const char *fmt, ...)
{
  int R;
  va_list ap;
  va_start(ap,fmt);
  R= fike_get_params(&cc->io, fmt, ap);
  va_end(ap);
  return R;
}
#endif

#ifdef FIKE_IMP_CLIENT
int fike_response_params(FIKE_SERVER *cc, const char *fmt, ...)
{
  int R;
  va_list ap;
  va_start(ap,fmt);
  R= fike_get_params(&cc->io, fmt, ap);
  va_end(ap);
  return R;
}

int fike_response_rest(FIKE_SERVER *cc, unsigned char **Rd, unsigned int *Rl)
{
  return fike_get_rest(&cc->io, Rd, Rl);
}
#endif

#ifdef FIKE_IMP_SERVER
int fike_command_rest(FIKE_CLIENT *cc, unsigned char **Rd, unsigned int *Rl)
{
  return fike_get_rest(&cc->io, Rd, Rl);
}
#endif

static void fike_remove_zabit(int type, void *data)
{
  int i;
  for(i=0;i<fike_n_zabit;i++)
    if (fike_zabit[i]->type==type && fike_zabit[i]->data==data)
    {
      fike_zabit[i]= fike_zabit[fike_n_zabit-1];
      fike_n_zabit--;
      i--;
    }
}

#ifdef FIKE_IMP_CLIENT
void fike_close_server(FIKE_SERVER *ss)
{
  fike_parray_remove_value(&fike_server_a, ss);
  fike_remove_zabit(FIKE_ZABIT_SERVER, ss);
  fike_close_sock(ss->io.fd);
  free(ss);
}
#endif

#ifdef FIKE_IMP_SERVER
void fike_close_client(FIKE_CLIENT *cc)
{
   fike_parray_remove_value(&fike_client_a, cc);
   fike_remove_zabit(FIKE_ZABIT_MUSTERI, cc);
   fike_close_sock(cc->io.fd);
   free(cc);
}
#endif

#ifdef fike_self_test
#include <ctype.h>


static int fike_prtc(int x)
{
  if (x>0x20 && x<0x7f) return x;
  return '.';
}

static void fike_print_message(unsigned char *buf, int rl)
{
  int r,c;
  printf("message received: %d bytes\n", rl);
  for(r=0;r<5;r++)
  {
    for(c=0;c<10;c++) printf("%02x ", buf[r*10+c]);
    for(c=0;c<10;c++) printf("%c", fike_prtc(buf[r*10+c]));
    printf("\n");
  }
}
#endif

void* __fike_unused_stuff[]= { fike_gint16 };


/***********************************************************************
 *                                                                     *
 *                    Notes from Old Implementation                    *
 *                                                                     *
 ***********************************************************************

 each command consists of:

  4 bytes magic
  4 bytes command id
  2 bytes data size
  data ..

  When a client becomes
  active, it can query the local network to find out active hosts.

  The query contains a globpattern. Each active host looks at this
  pattern and responds with a <name,ip,port> tuple if the pattern
  matches its name.

  For example, to find out all hosts a client will send a query
  which looks like this:

  Q(*)

 ***********************************************************************/
